@@ -17,10 +17,9 @@ structures.
 
 from __future__ import absolute_import, unicode_literals
 
-from mozbuild.util import (
-    shell_quote,
-    StrictOrderingOnAppendList,
-)
+from mozbuild.util import StrictOrderingOnAppendList
+from mozpack.chrome.manifest import ManifestEntry
+
 import mozpack.path as mozpath
 from .context import FinalTargetValue
 
@@ -35,6 +34,7 @@ from ..testing import (
 
 class TreeMetadata(object):
     """Base class for all data being captured."""
+    __slots__ = ()
 
 
 class ContextDerived(TreeMetadata):
@@ -45,13 +45,15 @@ class ContextDerived(TreeMetadata):
     """
 
     __slots__ = (
-        'objdir',
-        'relativedir',
+        'context_main_path',
         'context_all_paths',
-        'context_path',
-        'srcdir',
-        'topobjdir',
         'topsrcdir',
+        'topobjdir',
+        'relativedir',
+        'srcdir',
+        'objdir',
+        'config',
+        '_context',
     )
 
     def __init__(self, context):
@@ -77,11 +79,23 @@ class ContextDerived(TreeMetadata):
     def install_target(self):
         return self._context['FINAL_TARGET']
 
+    @property
+    def defines(self):
+        defines = self._context['DEFINES']
+        return Defines(self._context, defines) if defines else None
+
     @property
     def relobjdir(self):
         return mozpath.relpath(self.objdir, self.topobjdir)
 
 
+class HostMixin(object):
+    @property
+    def defines(self):
+        defines = self._context['HOST_DEFINES']
+        return HostDefines(self._context, defines) if defines else None
+
+
 class DirectoryTraversal(ContextDerived):
     """Describes how directory traversal for building should work.
 
@@ -96,14 +110,12 @@ class DirectoryTraversal(ContextDerived):
     """
     __slots__ = (
         'dirs',
-        'test_dirs',
     )
 
     def __init__(self, context):
         ContextDerived.__init__(self, context)
 
         self.dirs = []
-        self.test_dirs = []
 
 
 class BaseConfigSubstitution(ContextDerived):
@@ -127,10 +139,6 @@ class ConfigFileSubstitution(BaseConfigSubstitution):
     """Describes a config file that will be generated using substitutions."""
 
 
-class HeaderFileSubstitution(BaseConfigSubstitution):
-    """Describes a header file that will be generated using substitutions."""
-
-
 class VariablePassthru(ContextDerived):
     """A dict of variables to pass through to backend.mk unaltered.
 
@@ -150,9 +158,10 @@ class XPIDLFile(ContextDerived):
     """Describes an XPIDL file to be compiled."""
 
     __slots__ = (
-        'add_to_manifest',
-        'basename',
         'source_path',
+        'basename',
+        'module',
+        'add_to_manifest',
     )
 
     def __init__(self, context, source, module, add_to_manifest):
@@ -180,7 +189,7 @@ class BaseDefines(ContextDerived):
             elif value is False:
                 yield('-U%s' % define)
             else:
-                yield('-D%s=%s' % (define, shell_quote(value)))
+                yield('-D%s=%s' % (define, value))
 
     def update(self, more_defines):
         if isinstance(more_defines, Defines):
@@ -194,21 +203,6 @@ class Defines(BaseDefines):
 class HostDefines(BaseDefines):
     pass
 
-class Exports(ContextDerived):
-    """Context derived container object for EXPORTS, which is a
-    HierarchicalStringList.
-
-    We need an object derived from ContextDerived for use in the backend, so
-    this object fills that role. It just has a reference to the underlying
-    HierarchicalStringList, which is created when parsing EXPORTS.
-    """
-    __slots__ = ('exports', 'dist_install')
-
-    def __init__(self, context, exports, dist_install=True):
-        ContextDerived.__init__(self, context)
-        self.exports = exports
-        self.dist_install = dist_install
-
 class TestHarnessFiles(ContextDerived):
     """Sandbox container object for TEST_HARNESS_FILES,
     which is a HierarchicalStringList.
@@ -225,49 +219,6 @@ class TestHarnessFiles(ContextDerived):
         self.srcdir_pattern_files = srcdir_pattern_files
         self.objdir_files = objdir_files
 
-class Resources(ContextDerived):
-    """Context derived container object for RESOURCE_FILES, which is a
-    HierarchicalStringList, with an extra ``.preprocess`` property on each
-    entry.
-
-    The local defines plus anything in ACDEFINES are stored in ``defines`` as a
-    dictionary, for any files that need preprocessing.
-    """
-    __slots__ = ('resources', 'defines')
-
-    def __init__(self, context, resources, defines=None):
-        ContextDerived.__init__(self, context)
-        self.resources = resources
-        defs = {}
-        defs.update(context.config.defines)
-        if defines:
-            defs.update(defines)
-        self.defines = defs
-
-class BrandingFiles(ContextDerived):
-    """Sandbox container object for BRANDING_FILES, which is a
-    HierarchicalStringList.
-
-    We need an object derived from ContextDerived for use in the backend, so
-    this object fills that role. It just has a reference to the underlying
-    HierarchicalStringList, which is created when parsing BRANDING_FILES.
-    """
-    __slots__ = ('files')
-
-    def __init__(self, sandbox, files):
-        ContextDerived.__init__(self, sandbox)
-        self.files = files
-
-class JsPreferenceFile(ContextDerived):
-    """Context derived container object for a Javascript preference file.
-
-    Paths are assumed to be relative to the srcdir."""
-    __slots__ = ('path')
-
-    def __init__(self, context, path):
-        ContextDerived.__init__(self, context)
-        self.path = path
-
 class IPDLFile(ContextDerived):
     """Describes an individual .ipdl source file."""
 
@@ -375,7 +326,7 @@ class LinkageWrongKindError(Exception):
 class Linkable(ContextDerived):
     """Generic context derived container object for programs and libraries"""
     __slots__ = (
-        'defines',
+        'lib_defines',
         'linked_libraries',
         'linked_system_libs',
     )
@@ -384,7 +335,7 @@ class Linkable(ContextDerived):
         ContextDerived.__init__(self, context)
         self.linked_libraries = []
         self.linked_system_libs = []
-        self.defines = Defines(context, {})
+        self.lib_defines = Defines(context, {})
 
     def link_library(self, obj):
         assert isinstance(obj, BaseLibrary)
@@ -438,7 +389,7 @@ class Program(BaseProgram):
     KIND = 'target'
 
 
-class HostProgram(BaseProgram):
+class HostProgram(HostMixin, BaseProgram):
     """Context derived container object for HOST_PROGRAM"""
     SUFFIX_VAR = 'HOST_BIN_SUFFIX'
     KIND = 'host'
@@ -450,7 +401,7 @@ class SimpleProgram(BaseProgram):
     KIND = 'target'
 
 
-class HostSimpleProgram(BaseProgram):
+class HostSimpleProgram(HostMixin, BaseProgram):
     """Context derived container object for each program in
     HOST_SIMPLE_PROGRAMS"""
     SUFFIX_VAR = 'HOST_BIN_SUFFIX'
@@ -513,6 +464,7 @@ class SharedLibrary(Library):
     __slots__ = (
         'soname',
         'variant',
+        'symbols_file',
     )
 
     FRAMEWORK = 1
@@ -520,7 +472,7 @@ class SharedLibrary(Library):
     MAX_VARIANT = 3
 
     def __init__(self, context, basename, real_name=None, is_sdk=False,
-            soname=None, variant=None):
+            soname=None, variant=None, symbols_file=False):
         assert(variant in range(1, self.MAX_VARIANT) or variant is None)
         Library.__init__(self, context, basename, real_name, is_sdk)
         self.variant = variant
@@ -549,6 +501,13 @@ class SharedLibrary(Library):
         else:
             self.soname = self.lib_name
 
+        if not symbols_file:
+            self.symbols_file = None
+        elif context.config.substs['OS_TARGET'] == 'WINNT':
+            self.symbols_file = '%s.def' % self.lib_name
+        else:
+            self.symbols_file = '%s.symbols' % self.lib_name
+
 
 class ExternalLibrary(object):
     """Empty mixin for libraries built by an external build system."""
@@ -564,7 +523,7 @@ class ExternalSharedLibrary(SharedLibrary, ExternalLibrary):
     build system."""
 
 
-class HostLibrary(BaseLibrary):
+class HostLibrary(HostMixin, BaseLibrary):
     """Context derived container object for a host library"""
     KIND = 'host'
 
@@ -648,18 +607,6 @@ class LocalInclude(ContextDerived):
 
         self.path = path
 
-class GeneratedInclude(ContextDerived):
-    """Describes an individual generated include path."""
-
-    __slots__ = (
-        'path',
-    )
-
-    def __init__(self, context, path):
-        ContextDerived.__init__(self, context)
-
-        self.path = path
-
 
 class PerSourceFlag(ContextDerived):
     """Describes compiler flags specified for individual source files."""
@@ -693,21 +640,6 @@ class JARManifest(ContextDerived):
         self.path = path
 
 
-class JavaScriptModules(ContextDerived):
-    """Describes a JavaScript module."""
-
-    __slots__ = (
-        'modules',
-        'flavor',
-    )
-
-    def __init__(self, context, modules, flavor):
-        super(JavaScriptModules, self).__init__(context)
-
-        self.modules = modules
-        self.flavor = flavor
-
-
 class ContextWrapped(ContextDerived):
     """Generic context derived container object for a wrapped rich object.
 
@@ -785,7 +717,7 @@ class GeneratedSources(BaseSources):
         BaseSources.__init__(self, context, files, canonical_suffix)
 
 
-class HostSources(BaseSources):
+class HostSources(HostMixin, BaseSources):
     """Represents files to be compiled for the host during the build."""
 
     def __init__(self, context, files, canonical_suffix):
@@ -859,28 +791,59 @@ class FinalTargetFiles(ContextDerived):
     this object fills that role. It just has a reference to the underlying
     HierarchicalStringList, which is created when parsing FINAL_TARGET_FILES.
     """
-    __slots__ = ('files', 'target')
+    __slots__ = ('files')
 
-    def __init__(self, sandbox, files, target):
+    def __init__(self, sandbox, files):
         ContextDerived.__init__(self, sandbox)
         self.files = files
-        self.target = target
 
 
-class DistFiles(ContextDerived):
-    """Sandbox container object for FINAL_TARGET_FILES, which is a
+class FinalTargetPreprocessedFiles(ContextDerived):
+    """Sandbox container object for FINAL_TARGET_PP_FILES, which is a
     HierarchicalStringList.
 
     We need an object derived from ContextDerived for use in the backend, so
     this object fills that role. It just has a reference to the underlying
-    HierarchicalStringList, which is created when parsing DIST_FILES.
+    HierarchicalStringList, which is created when parsing
+    FINAL_TARGET_PP_FILES.
     """
-    __slots__ = ('files', 'target')
+    __slots__ = ('files')
 
-    def __init__(self, sandbox, files, target):
+    def __init__(self, sandbox, files):
         ContextDerived.__init__(self, sandbox)
         self.files = files
-        self.target = target
+
+
+class TestingFiles(FinalTargetFiles):
+    @property
+    def install_target(self):
+        return '_tests'
+
+
+class Exports(FinalTargetFiles):
+    """Context derived container object for EXPORTS, which is a
+    HierarchicalStringList.
+
+    We need an object derived from ContextDerived for use in the backend, so
+    this object fills that role. It just has a reference to the underlying
+    HierarchicalStringList, which is created when parsing EXPORTS.
+    """
+    @property
+    def install_target(self):
+        return 'dist/include'
+
+
+class BrandingFiles(FinalTargetFiles):
+    """Sandbox container object for BRANDING_FILES, which is a
+    HierarchicalStringList.
+
+    We need an object derived from ContextDerived for use in the backend, so
+    this object fills that role. It just has a reference to the underlying
+    HierarchicalStringList, which is created when parsing BRANDING_FILES.
+    """
+    @property
+    def install_target(self):
+        return 'dist/branding'
 
 
 class GeneratedFile(ContextDerived):
@@ -891,14 +854,16 @@ class GeneratedFile(ContextDerived):
         'method',
         'output',
         'inputs',
+        'flags',
     )
 
-    def __init__(self, context, script, method, output, inputs):
+    def __init__(self, context, script, method, output, inputs, flags=()):
         ContextDerived.__init__(self, context)
         self.script = script
         self.method = method
         self.output = output
         self.inputs = inputs
+        self.flags = flags
 
 
 class ClassPathEntry(object):
@@ -1012,3 +977,21 @@ class AndroidExtraPackages(ContextDerived):
     def __init__(self, context, packages):
         ContextDerived.__init__(self, context)
         self.packages = packages
+
+class ChromeManifestEntry(ContextDerived):
+    """Represents a chrome.manifest entry."""
+
+    __slots__ = (
+        'path',
+        'entry',
+    )
+
+    def __init__(self, context, manifest_path, entry):
+        ContextDerived.__init__(self, context)
+        assert isinstance(entry, ManifestEntry)
+        self.path = mozpath.join(self.install_target, manifest_path)
+        # Ensure the entry is relative to the directory containing the
+        # manifest path.
+        entry = entry.rebase(mozpath.dirname(manifest_path))
+        # Then add the install_target to the entry base directory.
+        self.entry = entry.move(mozpath.dirname(self.path))
