@@ -36,6 +36,7 @@ from io import StringIO
 
 from mozbuild.util import (
     EmptyValue,
+    HierarchicalStringList,
     memoize,
     ReadOnlyDefaultDict,
 )
@@ -133,6 +134,7 @@ class EmptyConfig(object):
         self.substs_unicode = self.PopulateOnGetDict(EmptyValue, udict)
         self.defines = self.substs
         self.external_source_dir = None
+        self.error_is_fatal = False
 
 
 def is_read_allowed(path, config):
@@ -219,6 +221,8 @@ class MozbuildSandbox(Sandbox):
         return Sandbox.__contains__(self, key)
 
     def __setitem__(self, key, value):
+        if key in self.special_variables and value is self[key]:
+            return
         if key in self.special_variables or key in self.functions or key in self.subcontext_types:
             raise KeyError('Cannot set "%s" because it is a reserved keyword'
                            % key)
@@ -322,7 +326,10 @@ class MozbuildSandbox(Sandbox):
         print('WARNING: %s' % message, file=sys.stderr)
 
     def _error(self, message):
-        raise SandboxCalledError(self._context.source_stack, message)
+        if self._context.error_is_fatal:
+            raise SandboxCalledError(self._context.source_stack, message)
+        else:
+            self._warning(message)
 
     def _template_decorator(self, func):
         """Registers a template function."""
@@ -413,7 +420,7 @@ class MozbuildSandbox(Sandbox):
             for key, value in context.items():
                 if isinstance(value, dict):
                     self[key].update(value)
-                elif isinstance(value, list):
+                elif isinstance(value, (list, HierarchicalStringList)):
                     self[key] += value
                 else:
                     self[key] = value
@@ -1124,13 +1131,8 @@ class BuildReader(object):
         # processing is performed.
         yield context
 
-        # We first collect directories populated in variables.
-        dir_vars = ['DIRS']
-
-        if context.config.substs.get('ENABLE_TESTS', False) == '1':
-            dir_vars.append('TEST_DIRS')
-
-        dirs = [(v, context[v]) for v in dir_vars if v in context]
+        # We need the list of directories pre-gyp processing for later.
+        dirs = list(context.get('DIRS', []))
 
         curdir = mozpath.dirname(path)
 
@@ -1180,20 +1182,19 @@ class BuildReader(object):
         # make backend needs order preserved. Once we autogenerate all backend
         # files, we should be able to convert this to a set.
         recurse_info = OrderedDict()
-        for var, var_dirs in dirs:
-            for d in var_dirs:
-                if d in recurse_info:
-                    raise SandboxValidationError(
-                        'Directory (%s) registered multiple times in %s' % (
-                            mozpath.relpath(d.full_path, context.srcdir), var),
-                        context)
+        for d in dirs:
+            if d in recurse_info:
+                raise SandboxValidationError(
+                    'Directory (%s) registered multiple times' % (
+                        mozpath.relpath(d.full_path, context.srcdir)),
+                    context)
 
-                recurse_info[d] = {}
-                for key in sandbox.metadata:
-                    if key == 'exports':
-                        sandbox.recompute_exports()
+            recurse_info[d] = {}
+            for key in sandbox.metadata:
+                if key == 'exports':
+                    sandbox.recompute_exports()
 
-                    recurse_info[d][key] = dict(sandbox.metadata[key])
+                recurse_info[d][key] = dict(sandbox.metadata[key])
 
         for path, child_metadata in recurse_info.items():
             child_path = path.join('moz.build').full_path
@@ -1315,7 +1316,7 @@ class BuildReader(object):
             # Explicitly set directory traversal variables to override default
             # traversal rules.
             if not isinstance(context, SubContext):
-                for v in ('DIRS', 'GYP_DIRS', 'TEST_DIRS'):
+                for v in ('DIRS', 'GYP_DIRS'):
                     context[v][:] = []
 
                 context['DIRS'] = sorted(dirs[context.main_path])
